// AnalogInputs.js
//  This module handles analog input processing and eventing

//#region Dependencies, Constants and Globals
//#region Dependencies
// Requires
var fs = require('fs');
var moment = require('moment');
var exec = require('child_process').exec;

var _this = this;// To allow access to exported functions
//#endregion Dependencies

//#region Constants
// Define the configuraton for the I/O
var ANALOG_INPUTS_LOCATION = '/sys/class/hwmon';
var ANALOG_INPUTS_CHIPS_PER_BOARD = 2;// Default = 2
var ANALOG_INPUTS_POINTS_PER_CHIP = 16;// Default = 16
var ANALOG_INPUTS_CYCLE_TIME = 750// Default = 500ms/750ms
var ANALOG_INPUTS_RATE = parseInt(ANALOG_INPUTS_CYCLE_TIME / (ANALOG_INPUTS_CHIPS_PER_BOARD * ANALOG_INPUTS_POINTS_PER_CHIP)*0.9)
var ANALOG_INPUTS_MAXIMUM_RANGE = 25;// Default = 25
var ANALOG_INPUTS_DEFAULT_LOW_THRESHOLD = 0;// Default = 0
var ANALOG_INPUTS_DEFAULT_HIGH_THRESHOLD = ANALOG_INPUTS_MAXIMUM_RANGE;// Default = ANALOG_INPUTS_MAXIMUM_RANGE
var ANALOG_INPUTS_RAWVALUE_OFFSET = -2.1796;// Default = -2.1796
var ANALOG_INPUTS_RAWVALUE_SLOPE = 0.0112;// Default = 0.0112
var ANALOG_INPUTS_CHANGE_FILTER = 5;// Default = ~50mV - this is the amount of change in the raw value we look at before we send socket update
var ANALOG_INPUTS_FIRST_READ = true;// Track the first read

//var TotalDuration = 0;
//var TotalReads = 0;

// These constants are for use as 
var ANALOG_INPUTS_IN_RANGE = 0;
var ANALOG_INPUTS_BELOW_THRESHOLD = 1;
var ANALOG_INPUTS_ABOVE_THRESHOLD = 2;
var ThresholdValues = ["NORMAL", "BELOW", "ABOVE"];

// SNMP - this is here for now, could move
var AuthProtocol = ["None", "MD5", "SHA"];
var PrivProtocol = ["None", "DES", "AES"];
//#endregion Constants

//#region Globals
// Timing analysis    
var totalCycleTime = 0.0;
var averageCycleTime = 0.0;
var totalCycles = 0;
var analogInputsReadStart;
var AnalogInputPointsRead = 0;

var SummitInfo = {};
var SmtpManager = {};
var SnmpManager = {};
var SqliteManager = {};
var LogManager = {};
var SocketIoManager = {};

// The following array holds, in memory, the current value (in raw bit values - 0 to 2500)
var analogInputsArray = [
    { "Address" : "0", "Name" : "Name", "FormulaId" : 0, "FormulaEnabled" : 1, "RawValue": 0, "RangedValue": 0 , "FormulaValue": 0, "ThresholdState": ANALOG_INPUTS_IN_RANGE, "TempValue": 0, "HoldTimerActive": false, "HoldTimer": 0, "Enabled": "FALSE", "SendEvents": "FALSE" , "Range": ANALOG_INPUTS_MAXIMUM_RANGE, "LowThreshold": 0, "HighThreshold": 0, "FormulaLowThreshold": 0, "FormulaHighThreshold": 0, "Offset": 0, "HoldTime": 0, "Qualifier": 0, "QualifierValue": 0, "TimeChanged": 0 }, 
    { "Address" : "1", "Name" : "Name", "FormulaId" : 0, "FormulaEnabled" : 1, "RawValue": 0, "RangedValue": 0 , "FormulaValue": 0, "ThresholdState": ANALOG_INPUTS_IN_RANGE, "TempValue": 0, "HoldTimerActive": false, "HoldTimer": 0, "Enabled": "FALSE", "SendEvents": "FALSE" , "Range": ANALOG_INPUTS_MAXIMUM_RANGE, "LowThreshold": 0, "HighThreshold": 0, "FormulaLowThreshold": 0, "FormulaHighThreshold": 0, "Offset": 0, "HoldTime": 0, "Qualifier": 0, "QualifierValue": 0, "TimeChanged": 0 }, 
    { "Address" : "2", "Name" : "Name", "FormulaId" : 0, "FormulaEnabled" : 1, "RawValue": 0, "RangedValue": 0 , "FormulaValue": 0, "ThresholdState": ANALOG_INPUTS_IN_RANGE, "TempValue": 0, "HoldTimerActive": false, "HoldTimer": 0, "Enabled": "FALSE", "SendEvents": "FALSE" , "Range": ANALOG_INPUTS_MAXIMUM_RANGE, "LowThreshold": 0, "HighThreshold": 0, "FormulaLowThreshold": 0, "FormulaHighThreshold": 0, "Offset": 0, "HoldTime": 0, "Qualifier": 0, "QualifierValue": 0, "TimeChanged": 0 }, 
    { "Address" : "3", "Name" : "Name", "FormulaId" : 0, "FormulaEnabled" : 1, "RawValue": 0, "RangedValue": 0 , "FormulaValue": 0, "ThresholdState": ANALOG_INPUTS_IN_RANGE, "TempValue": 0, "HoldTimerActive": false, "HoldTimer": 0, "Enabled": "FALSE", "SendEvents": "FALSE" , "Range": ANALOG_INPUTS_MAXIMUM_RANGE, "LowThreshold": 0, "HighThreshold": 0, "FormulaLowThreshold": 0, "FormulaHighThreshold": 0, "Offset": 0, "HoldTime": 0, "Qualifier": 0, "QualifierValue": 0, "TimeChanged": 0 }, 
    { "Address" : "4", "Name" : "Name", "FormulaId" : 0, "FormulaEnabled" : 1, "RawValue": 0, "RangedValue": 0 , "FormulaValue": 0, "ThresholdState": ANALOG_INPUTS_IN_RANGE, "TempValue": 0, "HoldTimerActive": false, "HoldTimer": 0, "Enabled": "FALSE", "SendEvents": "FALSE" , "Range": ANALOG_INPUTS_MAXIMUM_RANGE, "LowThreshold": 0, "HighThreshold": 0, "FormulaLowThreshold": 0, "FormulaHighThreshold": 0, "Offset": 0, "HoldTime": 0, "Qualifier": 0, "QualifierValue": 0, "TimeChanged": 0 }, 
    { "Address" : "5", "Name" : "Name", "FormulaId" : 0, "FormulaEnabled" : 1, "RawValue": 0, "RangedValue": 0 , "FormulaValue": 0, "ThresholdState": ANALOG_INPUTS_IN_RANGE, "TempValue": 0, "HoldTimerActive": false, "HoldTimer": 0, "Enabled": "FALSE", "SendEvents": "FALSE" , "Range": ANALOG_INPUTS_MAXIMUM_RANGE, "LowThreshold": 0, "HighThreshold": 0, "FormulaLowThreshold": 0, "FormulaHighThreshold": 0, "Offset": 0, "HoldTime": 0, "Qualifier": 0, "QualifierValue": 0, "TimeChanged": 0 }, 
    { "Address" : "6", "Name" : "Name", "FormulaId" : 0, "FormulaEnabled" : 1, "RawValue": 0, "RangedValue": 0 , "FormulaValue": 0, "ThresholdState": ANALOG_INPUTS_IN_RANGE, "TempValue": 0, "HoldTimerActive": false, "HoldTimer": 0, "Enabled": "FALSE", "SendEvents": "FALSE" , "Range": ANALOG_INPUTS_MAXIMUM_RANGE, "LowThreshold": 0, "HighThreshold": 0, "FormulaLowThreshold": 0, "FormulaHighThreshold": 0, "Offset": 0, "HoldTime": 0, "Qualifier": 0, "QualifierValue": 0, "TimeChanged": 0 }, 
    { "Address" : "7", "Name" : "Name", "FormulaId" : 0, "FormulaEnabled" : 1, "RawValue": 0, "RangedValue": 0 , "FormulaValue": 0, "ThresholdState": ANALOG_INPUTS_IN_RANGE, "TempValue": 0, "HoldTimerActive": false, "HoldTimer": 0, "Enabled": "FALSE", "SendEvents": "FALSE" , "Range": ANALOG_INPUTS_MAXIMUM_RANGE, "LowThreshold": 0, "HighThreshold": 0, "FormulaLowThreshold": 0, "FormulaHighThreshold": 0, "Offset": 0, "HoldTime": 0, "Qualifier": 0, "QualifierValue": 0, "TimeChanged": 0 }, 

    { "Address" : "8", "Name" : "Name", "FormulaId" : 0, "FormulaEnabled" : 1, "RawValue": 0, "RangedValue": 0 , "FormulaValue": 0, "ThresholdState": ANALOG_INPUTS_IN_RANGE, "TempValue": 0, "HoldTimerActive": false, "HoldTimer": 0, "Enabled": "FALSE", "SendEvents": "FALSE" , "Range": ANALOG_INPUTS_MAXIMUM_RANGE, "LowThreshold": 0, "HighThreshold": 0, "FormulaLowThreshold": 0, "FormulaHighThreshold": 0, "Offset": 0, "HoldTime": 0, "Qualifier": 0, "QualifierValue": 0, "TimeChanged": 0 }, 
    { "Address" : "9", "Name" : "Name", "FormulaId" : 0, "FormulaEnabled" : 1, "RawValue": 0, "RangedValue": 0 , "FormulaValue": 0, "ThresholdState": ANALOG_INPUTS_IN_RANGE, "TempValue": 0, "HoldTimerActive": false, "HoldTimer": 0, "Enabled": "FALSE", "SendEvents": "FALSE" , "Range": ANALOG_INPUTS_MAXIMUM_RANGE, "LowThreshold": 0, "HighThreshold": 0, "FormulaLowThreshold": 0, "FormulaHighThreshold": 0, "Offset": 0, "HoldTime": 0, "Qualifier": 0, "QualifierValue": 0, "TimeChanged": 0 }, 
    { "Address" : "10", "Name" : "Name", "FormulaId" : 0, "FormulaEnabled" : 1, "RawValue": 0, "RangedValue": 0 , "FormulaValue": 0, "ThresholdState": ANALOG_INPUTS_IN_RANGE, "TempValue": 0, "HoldTimerActive": false, "HoldTimer": 0, "Enabled": "FALSE", "SendEvents": "FALSE" , "Range": ANALOG_INPUTS_MAXIMUM_RANGE, "LowThreshold": 0, "HighThreshold": 0, "FormulaLowThreshold": 0, "FormulaHighThreshold": 0, "Offset": 0, "HoldTime": 0, "Qualifier": 0, "QualifierValue": 0, "TimeChanged": 0 }, 
    { "Address" : "11", "Name" : "Name", "FormulaId" : 0, "FormulaEnabled" : 1, "RawValue": 0, "RangedValue": 0 , "FormulaValue": 0, "ThresholdState": ANALOG_INPUTS_IN_RANGE, "TempValue": 0, "HoldTimerActive": false, "HoldTimer": 0, "Enabled": "FALSE", "SendEvents": "FALSE" , "Range": ANALOG_INPUTS_MAXIMUM_RANGE, "LowThreshold": 0, "HighThreshold": 0, "FormulaLowThreshold": 0, "FormulaHighThreshold": 0, "Offset": 0, "HoldTime": 0, "Qualifier": 0, "QualifierValue": 0, "TimeChanged": 0 }, 
    { "Address" : "12", "Name" : "Name", "FormulaId" : 0, "FormulaEnabled" : 1, "RawValue": 0, "RangedValue": 0 , "FormulaValue": 0, "ThresholdState": ANALOG_INPUTS_IN_RANGE, "TempValue": 0, "HoldTimerActive": false, "HoldTimer": 0, "Enabled": "FALSE", "SendEvents": "FALSE" , "Range": ANALOG_INPUTS_MAXIMUM_RANGE, "LowThreshold": 0, "HighThreshold": 0, "FormulaLowThreshold": 0, "FormulaHighThreshold": 0, "Offset": 0, "HoldTime": 0, "Qualifier": 0, "QualifierValue": 0, "TimeChanged": 0 }, 
    { "Address" : "13", "Name" : "Name", "FormulaId" : 0, "FormulaEnabled" : 1, "RawValue": 0, "RangedValue": 0 , "FormulaValue": 0, "ThresholdState": ANALOG_INPUTS_IN_RANGE, "TempValue": 0, "HoldTimerActive": false, "HoldTimer": 0, "Enabled": "FALSE", "SendEvents": "FALSE" , "Range": ANALOG_INPUTS_MAXIMUM_RANGE, "LowThreshold": 0, "HighThreshold": 0, "FormulaLowThreshold": 0, "FormulaHighThreshold": 0, "Offset": 0, "HoldTime": 0, "Qualifier": 0, "QualifierValue": 0, "TimeChanged": 0 }, 
    { "Address" : "14", "Name" : "Name", "FormulaId" : 0, "FormulaEnabled" : 1, "RawValue": 0, "RangedValue": 0 , "FormulaValue": 0, "ThresholdState": ANALOG_INPUTS_IN_RANGE, "TempValue": 0, "HoldTimerActive": false, "HoldTimer": 0, "Enabled": "FALSE", "SendEvents": "FALSE" , "Range": ANALOG_INPUTS_MAXIMUM_RANGE, "LowThreshold": 0, "HighThreshold": 0, "FormulaLowThreshold": 0, "FormulaHighThreshold": 0, "Offset": 0, "HoldTime": 0, "Qualifier": 0, "QualifierValue": 0, "TimeChanged": 0 }, 
    { "Address" : "15", "Name" : "Name", "FormulaId" : 0, "FormulaEnabled" : 1, "RawValue": 0, "RangedValue": 0 , "FormulaValue": 0, "ThresholdState": ANALOG_INPUTS_IN_RANGE, "TempValue": 0, "HoldTimerActive": false, "HoldTimer": 0, "Enabled": "FALSE", "SendEvents": "FALSE" , "Range": ANALOG_INPUTS_MAXIMUM_RANGE, "LowThreshold": 0, "HighThreshold": 0, "FormulaLowThreshold": 0, "FormulaHighThreshold": 0, "Offset": 0, "HoldTime": 0, "Qualifier": 0, "QualifierValue": 0, "TimeChanged": 0 }, 

    { "Address" : "16", "Name" : "Name", "FormulaId" : 0, "FormulaEnabled" : 1, "RawValue": 0, "RangedValue": 0 , "FormulaValue": 0, "ThresholdState": ANALOG_INPUTS_IN_RANGE, "TempValue": 0, "HoldTimerActive": false, "HoldTimer": 0, "Enabled": "FALSE", "SendEvents": "FALSE" , "Range": ANALOG_INPUTS_MAXIMUM_RANGE, "LowThreshold": 0, "HighThreshold": 0, "FormulaLowThreshold": 0, "FormulaHighThreshold": 0, "Offset": 0, "HoldTime": 0, "Qualifier": 0, "QualifierValue": 0, "TimeChanged": 0 }, 
    { "Address" : "17", "Name" : "Name", "FormulaId" : 0, "FormulaEnabled" : 1, "RawValue": 0, "RangedValue": 0 , "FormulaValue": 0, "ThresholdState": ANALOG_INPUTS_IN_RANGE, "TempValue": 0, "HoldTimerActive": false, "HoldTimer": 0, "Enabled": "FALSE", "SendEvents": "FALSE" , "Range": ANALOG_INPUTS_MAXIMUM_RANGE, "LowThreshold": 0, "HighThreshold": 0, "FormulaLowThreshold": 0, "FormulaHighThreshold": 0, "Offset": 0, "HoldTime": 0, "Qualifier": 0, "QualifierValue": 0, "TimeChanged": 0 }, 
    { "Address" : "18", "Name" : "Name", "FormulaId" : 0, "FormulaEnabled" : 1, "RawValue": 0, "RangedValue": 0 , "FormulaValue": 0, "ThresholdState": ANALOG_INPUTS_IN_RANGE, "TempValue": 0, "HoldTimerActive": false, "HoldTimer": 0, "Enabled": "FALSE", "SendEvents": "FALSE" , "Range": ANALOG_INPUTS_MAXIMUM_RANGE, "LowThreshold": 0, "HighThreshold": 0, "FormulaLowThreshold": 0, "FormulaHighThreshold": 0, "Offset": 0, "HoldTime": 0, "Qualifier": 0, "QualifierValue": 0, "TimeChanged": 0 }, 
    { "Address" : "19", "Name" : "Name", "FormulaId" : 0, "FormulaEnabled" : 1, "RawValue": 0, "RangedValue": 0 , "FormulaValue": 0, "ThresholdState": ANALOG_INPUTS_IN_RANGE, "TempValue": 0, "HoldTimerActive": false, "HoldTimer": 0, "Enabled": "FALSE", "SendEvents": "FALSE" , "Range": ANALOG_INPUTS_MAXIMUM_RANGE, "LowThreshold": 0, "HighThreshold": 0, "FormulaLowThreshold": 0, "FormulaHighThreshold": 0, "Offset": 0, "HoldTime": 0, "Qualifier": 0, "QualifierValue": 0, "TimeChanged": 0 }, 
    { "Address" : "20", "Name" : "Name", "FormulaId" : 0, "FormulaEnabled" : 1, "RawValue": 0, "RangedValue": 0 , "FormulaValue": 0, "ThresholdState": ANALOG_INPUTS_IN_RANGE, "TempValue": 0, "HoldTimerActive": false, "HoldTimer": 0, "Enabled": "FALSE", "SendEvents": "FALSE" , "Range": ANALOG_INPUTS_MAXIMUM_RANGE, "LowThreshold": 0, "HighThreshold": 0, "FormulaLowThreshold": 0, "FormulaHighThreshold": 0, "Offset": 0, "HoldTime": 0, "Qualifier": 0, "QualifierValue": 0, "TimeChanged": 0 }, 
    { "Address" : "21", "Name" : "Name", "FormulaId" : 0, "FormulaEnabled" : 1, "RawValue": 0, "RangedValue": 0 , "FormulaValue": 0, "ThresholdState": ANALOG_INPUTS_IN_RANGE, "TempValue": 0, "HoldTimerActive": false, "HoldTimer": 0, "Enabled": "FALSE", "SendEvents": "FALSE" , "Range": ANALOG_INPUTS_MAXIMUM_RANGE, "LowThreshold": 0, "HighThreshold": 0, "FormulaLowThreshold": 0, "FormulaHighThreshold": 0, "Offset": 0, "HoldTime": 0, "Qualifier": 0, "QualifierValue": 0, "TimeChanged": 0 }, 
    { "Address" : "22", "Name" : "Name", "FormulaId" : 0, "FormulaEnabled" : 1, "RawValue": 0, "RangedValue": 0 , "FormulaValue": 0, "ThresholdState": ANALOG_INPUTS_IN_RANGE, "TempValue": 0, "HoldTimerActive": false, "HoldTimer": 0, "Enabled": "FALSE", "SendEvents": "FALSE" , "Range": ANALOG_INPUTS_MAXIMUM_RANGE, "LowThreshold": 0, "HighThreshold": 0, "FormulaLowThreshold": 0, "FormulaHighThreshold": 0, "Offset": 0, "HoldTime": 0, "Qualifier": 0, "QualifierValue": 0, "TimeChanged": 0 }, 
    { "Address" : "23", "Name" : "Name", "FormulaId" : 0, "FormulaEnabled" : 1, "RawValue": 0, "RangedValue": 0 , "FormulaValue": 0, "ThresholdState": ANALOG_INPUTS_IN_RANGE, "TempValue": 0, "HoldTimerActive": false, "HoldTimer": 0, "Enabled": "FALSE", "SendEvents": "FALSE" , "Range": ANALOG_INPUTS_MAXIMUM_RANGE, "LowThreshold": 0, "HighThreshold": 0, "FormulaLowThreshold": 0, "FormulaHighThreshold": 0, "Offset": 0, "HoldTime": 0, "Qualifier": 0, "QualifierValue": 0, "TimeChanged": 0 }, 

    { "Address" : "24", "Name" : "Name", "FormulaId" : 0, "FormulaEnabled" : 1, "RawValue": 0, "RangedValue": 0 , "FormulaValue": 0, "ThresholdState": ANALOG_INPUTS_IN_RANGE, "TempValue": 0, "HoldTimerActive": false, "HoldTimer": 0, "Enabled": "FALSE", "SendEvents": "FALSE" , "Range": ANALOG_INPUTS_MAXIMUM_RANGE, "LowThreshold": 0, "HighThreshold": 0, "FormulaLowThreshold": 0, "FormulaHighThreshold": 0, "Offset": 0, "HoldTime": 0, "Qualifier": 0, "QualifierValue": 0, "TimeChanged": 0 }, 
    { "Address" : "25", "Name" : "Name", "FormulaId" : 0, "FormulaEnabled" : 1, "RawValue": 0, "RangedValue": 0 , "FormulaValue": 0, "ThresholdState": ANALOG_INPUTS_IN_RANGE, "TempValue": 0, "HoldTimerActive": false, "HoldTimer": 0, "Enabled": "FALSE", "SendEvents": "FALSE" , "Range": ANALOG_INPUTS_MAXIMUM_RANGE, "LowThreshold": 0, "HighThreshold": 0, "FormulaLowThreshold": 0, "FormulaHighThreshold": 0, "Offset": 0, "HoldTime": 0, "Qualifier": 0, "QualifierValue": 0, "TimeChanged": 0 }, 
    { "Address" : "26", "Name" : "Name", "FormulaId" : 0, "FormulaEnabled" : 1, "RawValue": 0, "RangedValue": 0 , "FormulaValue": 0, "ThresholdState": ANALOG_INPUTS_IN_RANGE, "TempValue": 0, "HoldTimerActive": false, "HoldTimer": 0, "Enabled": "FALSE", "SendEvents": "FALSE" , "Range": ANALOG_INPUTS_MAXIMUM_RANGE, "LowThreshold": 0, "HighThreshold": 0, "FormulaLowThreshold": 0, "FormulaHighThreshold": 0, "Offset": 0, "HoldTime": 0, "Qualifier": 0, "QualifierValue": 0, "TimeChanged": 0 }, 
    { "Address" : "27", "Name" : "Name", "FormulaId" : 0, "FormulaEnabled" : 1, "RawValue": 0, "RangedValue": 0 , "FormulaValue": 0, "ThresholdState": ANALOG_INPUTS_IN_RANGE, "TempValue": 0, "HoldTimerActive": false, "HoldTimer": 0, "Enabled": "FALSE", "SendEvents": "FALSE" , "Range": ANALOG_INPUTS_MAXIMUM_RANGE, "LowThreshold": 0, "HighThreshold": 0, "FormulaLowThreshold": 0, "FormulaHighThreshold": 0, "Offset": 0, "HoldTime": 0, "Qualifier": 0, "QualifierValue": 0, "TimeChanged": 0 }, 
    { "Address" : "28", "Name" : "Name", "FormulaId" : 0, "FormulaEnabled" : 1, "RawValue": 0, "RangedValue": 0 , "FormulaValue": 0, "ThresholdState": ANALOG_INPUTS_IN_RANGE, "TempValue": 0, "HoldTimerActive": false, "HoldTimer": 0, "Enabled": "FALSE", "SendEvents": "FALSE" , "Range": ANALOG_INPUTS_MAXIMUM_RANGE, "LowThreshold": 0, "HighThreshold": 0, "FormulaLowThreshold": 0, "FormulaHighThreshold": 0, "Offset": 0, "HoldTime": 0, "Qualifier": 0, "QualifierValue": 0, "TimeChanged": 0 }, 
    { "Address" : "29", "Name" : "Name", "FormulaId" : 0, "FormulaEnabled" : 1, "RawValue": 0, "RangedValue": 0 , "FormulaValue": 0, "ThresholdState": ANALOG_INPUTS_IN_RANGE, "TempValue": 0, "HoldTimerActive": false, "HoldTimer": 0, "Enabled": "FALSE", "SendEvents": "FALSE" , "Range": ANALOG_INPUTS_MAXIMUM_RANGE, "LowThreshold": 0, "HighThreshold": 0, "FormulaLowThreshold": 0, "FormulaHighThreshold": 0, "Offset": 0, "HoldTime": 0, "Qualifier": 0, "QualifierValue": 0, "TimeChanged": 0 }, 
    { "Address" : "30", "Name" : "Name", "FormulaId" : 0, "FormulaEnabled" : 1, "RawValue": 0, "RangedValue": 0 , "FormulaValue": 0, "ThresholdState": ANALOG_INPUTS_IN_RANGE, "TempValue": 0, "HoldTimerActive": false, "HoldTimer": 0, "Enabled": "FALSE", "SendEvents": "FALSE" , "Range": ANALOG_INPUTS_MAXIMUM_RANGE, "LowThreshold": 0, "HighThreshold": 0, "FormulaLowThreshold": 0, "FormulaHighThreshold": 0, "Offset": 0, "HoldTime": 0, "Qualifier": 0, "QualifierValue": 0, "TimeChanged": 0 }, 
    { "Address" : "31", "Name" : "Name", "FormulaId" : 0, "FormulaEnabled" : 1, "RawValue": 0, "RangedValue": 0 , "FormulaValue": 0, "ThresholdState": ANALOG_INPUTS_IN_RANGE, "TempValue": 0, "HoldTimerActive": false, "HoldTimer": 0, "Enabled": "FALSE", "SendEvents": "FALSE" , "Range": ANALOG_INPUTS_MAXIMUM_RANGE, "LowThreshold": 0, "HighThreshold": 0, "FormulaLowThreshold": 0, "FormulaHighThreshold": 0, "Offset": 0, "HoldTime": 0, "Qualifier": 0, "QualifierValue": 0, "TimeChanged": 0 }
];

//  Formulas Array
var FormulasArray = [];
//#endregion Globals
//#endregion Dependencies, Constants and Globals

//#region Analog Inputs Initialization
// Initialize the analog inputs upon system start
//      - Checks if database table exists, if not creates the table and fills it with default data for each point
//      - Else reads from the database and updates analogInputsArray
exports.InitializeAnalogInputs = function (info, smtp, snmp, db, io, logger, callback) {
    // Initialize globals
    SummitInfo = info;
    SmtpManager = smtp;
    SnmpManager = snmp;
    SqliteManager = db;
    SocketIoManager = io;
    LogManager = logger;
    
    // Try getting the AnalogInputs table
    SqliteManager.get("SELECT name FROM sqlite_master WHERE type='table' AND name='AnalogInputs'", function (err, rows) {
        if (err !== null) {
            LogManager.error('Analog Inputs - Initialize Error (Check)=' + JSON.stringify(err));
        }
        else if (typeof rows === "undefined") {
            // We read nothing, so let's create the table ...
            LogManager.info("Analog Inputs - Initialize by creating database table ...");
            SqliteManager.exec("CREATE TABLE if not exists AnalogInputs (Address INTEGER, Name TEXT, FormulaId INTEGER, FormulaEnabled INTEGER, RawValue INTEGER, RangedValue TEXT, ThresholdState INTEGER, Enabled TEXT, SendEvents TEXT, Range TEXT, LowThreshold TEXT, HighThreshold TEXT, Offset TEXT, HoldTime INTEGER, Qualifier INTEGER, TimeChanged TEXT )", function (err) {
                if (err !== null) {
                    LogManager.error('Analog Inputs - Initialize Error (Create)=' + JSON.stringify(err));
                }
                else {
                    var timeChanged = moment();// Record the create time as the last time the I/O point value was changed
                    
                    // Created the table, now add data ...
                    SqliteManager.run("BEGIN")
                    for (var i = 0; i < (ANALOG_INPUTS_CHIPS_PER_BOARD * ANALOG_INPUTS_POINTS_PER_CHIP); i++) {
                        // Initialize the array
                        UpdateAnalogInputConfiguration(i, "Analog Input " + (i + 1), -1, 1, 0, 0, 0, 0, "FALSE", "FALSE", ANALOG_INPUTS_MAXIMUM_RANGE, ANALOG_INPUTS_DEFAULT_LOW_THRESHOLD, ANALOG_INPUTS_DEFAULT_HIGH_THRESHOLD, 0, 0, 0, 0, 0, timeChanged);
                        analogInputsArray[i].RawValue = -1;// Force a new read
                        analogInputsArray[i].TempValue = -1;// Historical NOT USED anymore
                        analogInputsArray[i].HoldTimerActive = false;
                        analogInputsArray[i].HoldTimer = timeChanged;
                        
                        // Initialize the database
                        insertStatement = i + ", 'Analog Input " + (i + 1) + "', -1, 1, 0, '0', 0, 'FALSE', 'FALSE', '" + ANALOG_INPUTS_MAXIMUM_RANGE + "', '" + ANALOG_INPUTS_DEFAULT_LOW_THRESHOLD + "', '" + ANALOG_INPUTS_DEFAULT_HIGH_THRESHOLD + "', '0', 0, 0, '" + timeChanged + "'";
                        //logger.debug('INSERT INTO AnalogInputs (Address, Name, FormulaId, FormulaEnabled, RawValue, RangedValue, FormulaValue, Enabled, SendEvents, Range, LowThreshold, HighThreshold, FormulaLowThreshold, FormulaHighThreshold, Offset, HoldTime, Qualifier, TimeChanged) VALUES (' + insertStatement + ');');
                        SqliteManager.run('INSERT INTO AnalogInputs (Address, Name, FormulaId, FormulaEnabled, RawValue, RangedValue, ThresholdState, Enabled, SendEvents, Range, LowThreshold, HighThreshold, Offset, HoldTime, Qualifier, TimeChanged) VALUES (' + insertStatement + ');', function (err) {
                            if (err !== null) {
                                LogManager.error('Analog Inputs - Initialize Error (Insert)=' + JSON.stringify(err));
                            }
                        });
                    }
                    SqliteManager.run("COMMIT");
                    
                    // Analog Inputs table created
                    if (typeof callback == "function") {
                        LogManager.info('Analog Inputs - Initialized new database table');
                        callback();
                    }
                }
            });
        }
        else {
            // Write out current AnalogInputs table
            LogManager.info("Analog Inputs - Initialize by reading database ...");
            SqliteManager.each("SELECT rowid as id, Address, Name, FormulaId, FormulaEnabled, RawValue, RangedValue, ThresholdState, Enabled, SendEvents, Range, LowThreshold, HighThreshold, Offset, HoldTime, Qualifier, TimeChanged FROM AnalogInputs ORDER BY Address", function (err, row) {
                if (err !== null) {
                    LogManager.error('Analog Inputs - Initialize Error (Select)=' + JSON.stringify(err));
                }
                else {
                    row.RawValue = -1;// Force a new value read on startup
                    //row.ThresholdState = -1;// Force an evaluation of threshold on startup
                    UpdateAnalogInputConfiguration(row.Address, row.Name, row.FormulaId, row.FormulaEnabled, row.RawValue, row.RangedValue, 0, row.ThresholdState, row.Enabled, row.SendEvents, row.Range, row.LowThreshold, row.HighThreshold, 0, 0, row.Offset, row.HoldTime, row.Qualifier, row.TimeChanged)
                    LogManager.debug("Analog Inputs - [Value INITIAL from dB] Analog Input[" + (parseInt(row.Address) + 1) + "] '" + row.Name + "' RawValue=" + row.RawValue);
                }
            }, function () {
                // Analog Inputs read from table
                if (typeof callback == "function") {
                    LogManager.info('Analog Inputs - Initialized from existing database table');
                    callback();
                }
            });
        }
    });
}

// Adjust the analogInputsArray
//  This is only called during initialization of the dB and if the client makes a change
module.exports.UpdateAnalogInputConfiguration = UpdateAnalogInputConfiguration;
function UpdateAnalogInputConfiguration(Address, Name, FormulaId, FormulaEnabled, RawValue, RangedValue, FormulaValue, ThresholdState, Enabled, SendEvents, Range, LowThreshold, HighThreshold, FormulaLowThreshold, FormulaHighThreshold, Offset, HoldTime, Qualifier, TimeChanged) {
   
    analogInputsArray[Address].Name = Name;
    analogInputsArray[Address].FormulaId = FormulaId;
    analogInputsArray[Address].FormulaEnabled = FormulaEnabled;
    analogInputsArray[Address].RawValue = RawValue;
    analogInputsArray[Address].FormulaValue = FormulaValue;
    analogInputsArray[Address].RangedValue = RangedValue;
    analogInputsArray[Address].ThresholdState = ThresholdState;
    analogInputsArray[Address].Enabled = Enabled;
    analogInputsArray[Address].SendEvents = SendEvents;
    analogInputsArray[Address].Range = Range;
    analogInputsArray[Address].LowThreshold = LowThreshold;
    analogInputsArray[Address].HighThreshold = HighThreshold;
    analogInputsArray[Address].Offset = Offset;
    analogInputsArray[Address].HoldTime = HoldTime;
    analogInputsArray[Address].Qualifier = Qualifier;
    analogInputsArray[Address].QualifierValue = 0;
    analogInputsArray[Address].TempValue = -1;// Set to last known RawValue, will be fixed by first read
    analogInputsArray[Address].TimeChanged = TimeChanged;
    
    // Update FormulaValue, FormulaLowThreshold, FormulaHighThreshold info, if present
    var FormulaIndex = -1;
    if (analogInputsArray[Address].FormulaId !== -1) {
        //logger.info("UpdateAnalogInputConfiguration - Address=" + Address + " has a FormulaID=" + analogInputsArray[Address].FormulaId);
        // Find formula
        //logger.info("UpdateAnalogInputConfiguration - FormulasArray[" + f + "].length=" + FormulasArray.length);
        for (var f = 0; f < FormulasArray.length; f++) {
            if (analogInputsArray[Address].FormulaId === FormulasArray[f].Id) {
                FormulaIndex = f;
                break;
            }
        }
        //logger.info("UpdateAnalogInputConfiguration - FormulaIndex=" + FormulaIndex);
        
        // Check to see if have a formula, if not don't worry about FormulaValue for this change
        if (FormulaIndex !== -1) {
            analogInputsArray[Address].FormulaValue = _this.CalculateFormula(analogInputsArray[Address].RangedValue, FormulasArray[FormulaIndex]);
            //logger.info("UpdateAnalogInputConfiguration - Calculating RangedValue=" + analogInputsArray[Address].RangedValue + " -> FormulaValue=" + analogInputsArray[Address].FormulaValue);
            analogInputsArray[Address].FormulaLowThreshold = _this.CalculateFormula(analogInputsArray[Address].LowThreshold, FormulasArray[FormulaIndex]);
            //logger.info("UpdateAnalogInputConfiguration - Calculating LowThreshold=" + analogInputsArray[Address].LowThreshold + " -> FormulaLowThreshold=" + analogInputsArray[Address].FormulaLowThreshold);
            analogInputsArray[Address].FormulaHighThreshold = _this.CalculateFormula(analogInputsArray[Address].HighThreshold, FormulasArray[FormulaIndex]);
            //logger.info("UpdateAnalogInputConfiguration - Calculating HighThreshold=" + analogInputsArray[Address].HighThreshold + " -> FormulaHighThreshold=" + analogInputsArray[Address].FormulaHighThreshold);
        }
    } else {
        // Update based on what was sent in
        analogInputsArray[Address].FormulaValue = FormulaValue;
        analogInputsArray[Address].FormulaLowThreshold = FormulaLowThreshold;
        analogInputsArray[Address].FormulaHighThreshold = FormulaHighThreshold;
    }
    
    //logger.info("UpdateAnalogInputConfiguration - FormulaValue=" + analogInputsArray[Address].FormulaValue);
    //logger.info("UpdateAnalogInputConfiguration - FormulaLowThreshold=" + analogInputsArray[Address].FormulaLowThreshold);
    //logger.info("UpdateAnalogInputConfiguration - FormulaHighThreshold=" + analogInputsArray[Address].FormulaHighThreshold)
    
    LogManager.debug("UpdateAnalogInputConfiguration - analogInputsArray[" + (Address + 1) + "]=" + JSON.stringify(analogInputsArray[Address]));
    
    sendValueAnalogInputsConfigured();// Update the Analog Inputs configuration
}

exports.InitializeVoltageFormula = function (voltageFormulae,logger) {
    logger.info("analoginputs.InitializeVoltageFormulae: Setting voltage formulae based on Summit configuration=" + JSON.stringify(voltageFormulae));
    ANALOG_INPUTS_RAWVALUE_SLOPE = Number(voltageFormulae.slope);
    ANALOG_INPUTS_RAWVALUE_OFFSET = Number(voltageFormulae.offset);
    for (var a = 0; a < (ANALOG_INPUTS_CHIPS_PER_BOARD * ANALOG_INPUTS_POINTS_PER_CHIP); a++) {
        ForceNewRawValue(a);
    }
}
//#endregion Analog Inputs Initialization

//#region Analog Inputs API
exports.GetAnalogInputValues = function(address) {
    //Logger.debug("GetAnalogInputValues - address=" + address + " values=" + JSON.stringify(analogInputsArray[address]));
    return analogInputsArray[address];
}

module.exports.ConvertRawValueToValue = ConvertRawValueToValue;
function ConvertRawValueToValue(rawValue) {
    var value = Number(parseFloat(rawValue) * ANALOG_INPUTS_RAWVALUE_SLOPE + ANALOG_INPUTS_RAWVALUE_OFFSET).toFixed(2);
    if (value < 0) value = 0;
    LogManager.trace("ConvertRawValueToValue: rawValue=" + rawValue + "  value=" + value);
    return value;
}

module.exports.ConvertValueToRawValue = ConvertValueToRawValue;
function ConvertValueToRawValue(value) {
    var rawValue = Math.round((parseFloat(value) - ANALOG_INPUTS_RAWVALUE_OFFSET) / ANALOG_INPUTS_RAWVALUE_SLOPE);
    LogManager.trace("ConvertValueToRawVaule: value=" + value + "  rawValue=" + rawValue);
    return rawValue;
}

exports.UpdateQualifierValues = function(digitalInputsArray) {
    for (qualifier = 0; qualifier < digitalInputsArray.length; qualifier++) {
        for (analogInput = 0; analogInput < analogInputsArray.length ; analogInput++) {
            if (analogInputsArray[analogInput].Qualifier > 0) {
                // This analog has a qualifier defined, so let's update the current qualifier value
                if ((analogInputsArray[analogInput].Qualifier - 1) == qualifier) {
                    if (analogInputsArray[analogInput].QualifierValue != digitalInputsArray[qualifier].Value) {
                        LogManager.info("Analog Inputs - [REAL-TIME UPDATE(Config) - Qualifier Changed] Analog Input[" + (analogInput + 1) + "] Qualifier='" + digitalInputsArray[qualifier].Name + "' Value=" + digitalInputsArray[qualifier].Value);
                        sendValueAnalogInput("AnalogInput", analogInputsArray[analogInput]);// Updated qualifier value
                        if (digitalInputsArray[qualifier].Value == 1)
                            SocketIoManager.sockets.emit("AnalogInputQualifier", { "AnalogInput": analogInput, "Qualifier": true });
                        else
                            SocketIoManager.sockets.emit("AnalogInputQualifier", { "AnalogInput": analogInput, "Qualifier": false });
                    }
                    analogInputsArray[analogInput].QualifierValue = digitalInputsArray[qualifier].Value;
                }
            }
        }
    }
}

exports.GetAnalogInputCurrentValue = function(address) {
    //Logger.debug("GetAnalogInputCurrentValue - address=" + address + " value=" + analogInputsArray[address].RawValue);
    return analogInputsArray[address].RawValue;
}

exports.GetAnalogInputQualifierValue = function(address) {
    return analogInputsArray[address].QualifierValue;
}

module.exports.GetAnalogInputThresholdState = GetAnalogInputThresholdState;
function GetAnalogInputThresholdState(AnalogInput, Value) {
    LogManager.trace("Analog Inputs - GetAnalogInputThresholdState: AnalogInput " + AnalogInput.Address + " Value=" + Value);
    var LowThresholdRawValue = ConvertValueToRawValue(AnalogInput.LowThreshold);
    var HighThresholdRawValue = ConvertValueToRawValue(AnalogInput.HighThreshold);
    var CurrentRangedValue = ConvertRawValueToValue(Value);
    
    var ThresholdState = ANALOG_INPUTS_IN_RANGE;
    if (Value < LowThresholdRawValue) {
        // This specifically tests for the 0.00V case
        if (AnalogInput.LowThreshold != CurrentRangedValue) {
            ThresholdState = ANALOG_INPUTS_BELOW_THRESHOLD
        } else {
            //Logger.info("Analog Inputs - [NO THRESHOLD CHANGE] Analog Input[" + AnalogInput.Address + "] LowThreshold=" + AnalogInput.LowThreshold + " Value=" + CurrentRangedValue);
        }
    } else if (Value > HighThresholdRawValue) {
        ThresholdState = ANALOG_INPUTS_ABOVE_THRESHOLD
    }
    
    LogManager.trace("Analog Inputs - GetAnalogInputThresholdState: ThresholdState=" + ThresholdState);
    return ThresholdState;
}

// Use this function to force a read of the I/O
//  We need to do this after we change a configuration
//  We force both a read of the RawValue
//  160812: I took out calculation of ThresholdState so that we don't set off unncessary send out
//  This function basically forces a new evaulation, as a result socket and SNMP messages
module.exports.ForceNewRawValue = ForceNewRawValue;
function ForceNewRawValue(index) {
    //Logger.debug("ForceNewRawValue: For address index=" + index);
    analogInputsArray[index].RawValue = -1;
    analogInputsArray[index].ThresholdState = -1;// In case the threshold state was changed
}
//#endregion Analog Inputs API

//#region Analog Inputs Processing
// Intervalled processing of analog inputs
var processCounter = 0;
var processTotalTime = 0;
var startAnalogInputReadTime = 0;
var counterAnalogInputReadTime = 0;
var totalAnalogInputReadTime = 0;

// These variable hand interim real-time updates
var LastTimeBeforeInterimUpdate = moment();
var CurrentTimeBeforeInterimUpdate;
var ANALOGINPUTS_INTERIM_UPDATE = ANALOG_INPUTS_CYCLE_TIME * 20;// Make it ~15 seconds per interval (vs. 7.5sec previously)

exports.ProcessAnalogInputs = function (ledManager) {
    
    //var timeStart = moment();
    //logger.info("ProcessAnalogInputs called at "+moment().format("hh:mm:ss.SSS"))
    
    // Code to measure time between Analog I/O process calls
    //#region Analog I/O Read Cycle Process Calls
    //totalAnalogInputReadTime += timeStart - startAnalogInputReadTime;//How much time has elapsed since last time we were here
    //counterAnalogInputReadTime++;
    //startAnalogInputReadTime = timeStart;// Last time we started
    //if (counterAnalogInputReadTime === 10) {
    //    logger.info("ProcessAnalogInputs: Average time between call= " + Number((totalAnalogInputReadTime / counterAnalogInputReadTime)).toFixed(2) + "msec");
    //    totalAnalogInputReadTime = 0;
    //    counterAnalogInputReadTime = 0;
    //}
    //#endregion Analog I/O Read Cycle Process Calls
    
    if (ANALOG_INPUTS_FIRST_READ) {
        LogManager.trace("Analog Inputs - Starting processing at " + ANALOG_INPUTS_RATE + "ms/point");
    }
    
    var count = 0;
    var index = 0;
    AnalogInputPointsRead = 0;
    var chipCounter = 0;
    var inputCounter = 0;

    var analoginputloop = setInterval(function () {
        readAnalogInputFile(chipCounter, inputCounter)
        inputCounter++
        if (inputCounter == ANALOG_INPUTS_POINTS_PER_CHIP) {
            inputCounter = 0;
            chipCounter++;
            if (chipCounter == ANALOG_INPUTS_CHIPS_PER_BOARD) {
                clearInterval(analoginputloop);// We're done
                ANALOG_INPUTS_FIRST_READ = false;
                
                // To do a 15second real-time update, in case they've refreshed the screen
                CurrentTimeBeforeInterimUpdate = moment();
                if ((CurrentTimeBeforeInterimUpdate - LastTimeBeforeInterimUpdate) > ANALOGINPUTS_INTERIM_UPDATE) {
                    sendInterimAnalogInputs();// Interim analog values
                    LastTimeBeforeInterimUpdate = CurrentTimeBeforeInterimUpdate;
                } 
            }
        }
    }, ANALOG_INPUTS_RATE);
    
    //for (var chipCounter = 0; chipCounter < ANALOG_INPUTS_CHIPS_PER_BOARD ; chipCounter++) {
    //    for (var inputCounter = 0; inputCounter < ANALOG_INPUTS_POINTS_PER_CHIP; inputCounter++) {
    //        readAnalogInputFile(chipCounter, inputCounter, db, io, logger, netSnmp)
    //    }
    //}

    // Code to measure complete analog I/O read cycle
    //#region Complete Analog I/O Read Cycle
    //if (processCounter === 10) {
    //    logger.info("ProcessAnalogInputs: Average process time = " + Number((processTotalTime / processCounter)).toFixed(2) + "msec");
    //    processCounter = 0;
    //    processTotalTime = 0;
    //}
    //processCounter++;
    //processTotalTime += (moment() - timeStart);
    //#endregion Complete Analog I/O Read Cycle
}

function readAnalogInputFile(chipCounter, inputCounter) {
    var index = inputCounter + ANALOG_INPUTS_POINTS_PER_CHIP * chipCounter;
    var timeChanged = moment();
    
    //logger.info("Reading File: " + ANALOG_INPUTS_LOCATION + "/hwmon" + chipCounter + "/in" + inputCounter + "_input - Analog Input[" + index + "].Enabled=" + analogInputsArray[index].Enabled);
    
    // Only read value if the I/O point is enabled
    if (analogInputsArray[index].Enabled === "TRUE") {
        try {
            //timeStart = moment();
            var readValue = 0;
            var data = 0;
            try {
                data = fs.readFileSync(ANALOG_INPUTS_LOCATION + "/hwmon" + chipCounter + "/in" + inputCounter + "_input", 'utf8');
            } catch (err) {
                // This is to catch infrequent errors related to reading
                LogManager.error('Analog Inputs - Process Error (File Read): error=' + err + ' (Analog Input[' + (index + 1) + '] @ Location: Chip=' + chipCounter + ', Input=' + inputCounter + ')');
                return;
            }
            
            // Get rid of any offending extra characters in read
            readValue = String(data).replace(/(\r\n|\n|\r|[ +])/img, "");
            //logger.info("Read    File: " + ANALOG_INPUTS_LOCATION + "/hwmon" + chipCounter + "/in" + inputCounter + "_input - readValue=" + readValue);
            //logger.trace('Analog Inputs - Process OK (File Read)= Normal (Analog Input at Address=' + index + ' readValue=' + readValue + ') - Stored RawValue=' + analogInputsArray[index].RawValue);
           
            
            // Add the offset to the read value
            readValue = Math.round(Number(readValue) + Number(ConvertValueToRawValue(analogInputsArray[index].Offset)) - Number(ConvertValueToRawValue(0)));
            
            // Create temp value to track allow us to handle hold times & qualifiers
            tempAnalogInputPoint = JSON.parse(JSON.stringify(analogInputsArray[index]));
            tempAnalogInputPoint.RawValue = readValue;
            tempAnalogInputPoint.RangedValue = ConvertRawValueToValue(readValue);
            LogManager.trace("Analog Inputs - readAnalogInputFile: Read Logic - tempAnalogInputPoint=" + JSON.stringify(tempAnalogInputPoint));
            
            // Filter logic
            //  - We ignore minute changes in the analog values, 
            //    change in raw value must be greater than ANALOG_INPUTS_CHANGE_FILTER
            //  - But we need to go through the logic when a HoldTimer is being used
            //#region Filter logic
            LogManager.trace("Analog Inputs - readAnalogInputFile: Filter Logic");
            // Don't do anything if:
            //  - there has been no change in the raw value reading (so there's no real change in the input)
            //  - there is no hold time (so we don't have to start any timers)
            //  - the threshold state is valid (it hasn't been invalidated because of a configuration change)
            var valueChanged = Math.abs(analogInputsArray[index].RawValue - readValue);
            if ((valueChanged <= ANALOG_INPUTS_CHANGE_FILTER) 
                && (analogInputsArray[index].HoldTime == 0) 
                && (analogInputsArray[index].ThresholdState >= 0)) {
                return;
            }

            //if (valueChanged > ANALOG_INPUTS_CHANGE_FILTER) logger.info("Value Changed: " + ANALOG_INPUTS_LOCATION + "/hwmon" + chipCounter + "/in" + inputCounter + "_input - readValue=" + readValue + " previous RawValue=" + analogInputsArray[index].RawValue + " valueChanged=" + valueChanged);

            if (analogInputsArray[index].HoldTime > 0) {
                LogManager.trace("Analog Inputs - [Filter: HoldTimer Used] Analog Input[" + (index + 1) + "] old value[array]=" + analogInputsArray[index].RawValue + " new value[read]=" + readValue);
            } else {
                LogManager.trace("Analog Inputs - [Filter: Value Changed] Analog Input[" + (index + 1) + "] old value[array]=" + analogInputsArray[index].RawValue + " new value[read]=" + readValue);
            }
            //#endregion Filter logic
            
            // Threshold Change logic
            //  - ThresholdChanged = true or false, depending if we've changed a threshold
            //#region Threshold logic
            LogManager.trace("Analog Inputs - readAnalogInputFile: Threshold Logic");
            var ThresholdChanged = false;
            var oldThreshold = analogInputsArray[index].ThresholdState;
            var newThreshold = GetAnalogInputThresholdState(tempAnalogInputPoint, tempAnalogInputPoint.RawValue);
            if ((oldThreshold != newThreshold) || (oldThreshold < 0)) {
                ThresholdChanged = true;
                LogManager.trace("Analog Inputs - [Threshold=" + ThresholdChanged + "] Analog Input[" + (index + 1) + "] Stored=" + analogInputsArray[index].RangedValue + "V (" + (oldThreshold != -1?ThresholdValues[oldThreshold]:"UNDEFINED") + "),  RealTime=" + tempAnalogInputPoint.RangedValue + "V (" + (newThreshold != -1?ThresholdValues[newThreshold]:"UNDEFINED") + ")");
            } else {
                LogManager.trace("Analog Inputs - [Threshold=" + ThresholdChanged + "] Analog Input[" + (index + 1) + "] Stored=" + analogInputsArray[index].RangedValue + "V (" + (oldThreshold != -1?ThresholdValues[oldThreshold]:"UNDEFINED") + "),  RealTime=" + tempAnalogInputPoint.RangedValue + "V (" + (newThreshold != -1?ThresholdValues[newThreshold]:"UNDEFINED") + ")");
            }
            //#endregion Threshold logic
            
            // Qualification logic
            //  - QualifierValid = true or false, depending if qualified by digital input
            //  - AnalogInput[index].HoldTimerActive = true or false, depending on qualifications status
            //  - AnalogInput[index].HoldTimer = holds time of when first qualified, with or without qualifier
            //#region Qualification logic
            LogManager.trace("Analog Inputs - readAnalogInputFile: Qualification Logic");
            var QualifierValid = false;
            
            if (analogInputsArray[index].Qualifier > 0) {
                // Qualifier USED logic
                //#region Qualifier USED logic
                if (analogInputsArray[index].QualifierValue > 0) {
                    //#region HoldTime logic for Qualifier USED & TRUE
                    QualifierValid = true;
                    //logger.trace("Analog Inputs - [Qualifier=USED, QualifierValid=TRUE] Analog Input[" + (index+1) + "] using qualifier=DigitalInput[" + (analogInputsArray[index].Qualifier) + "]");

                    if (analogInputsArray[index].HoldTime > 0) {
                        // If there is a hold time ...
                        if (analogInputsArray[index].HoldTimerActive === false) {
                            // If the hold timer was previously not running ...
                            if (ThresholdChanged) {
                                //  And the threshold changed ... start the timer
                                analogInputsArray[index].HoldTimer = timeChanged;
                                analogInputsArray[index].HoldTimerActive = true;
                                LogManager.info("Analog Inputs - [HOLD TIMER - HoldTimer=" + (analogInputsArray[index].HoldTime > 0?"STARTED (Threshold Change)":"NOT_USED") + "Qualifier=USED, QualifierValid=TRUE] Analog Input[" + (index + 1) + "] '" + analogInputsArray[index].Name + "' using qualifier=DigitalInput[" + (analogInputsArray[index].Qualifier) + "]");
                                SocketIoManager.sockets.emit("AnalogInputHoldTime", { "AnalogInput": index, "HoldTime": "HOLDING" })
                            }
                        } 
                        else {
                            // If the hold timer was previously running ...
                            if (!ThresholdChanged) {
                                // If the threshold changed ...
                                if (analogInputsArray[index].HoldTimerActive == true) {
                                    //  And the threshold has returned to same place ... reset the timer
                                    LogManager.info("Analog Inputs - [HOLD TIMER - HoldTimer=" + (analogInputsArray[index].HoldTime > 0?"RESET (Threshold Return)":"NOT_USED") + ", Qualifier=USED, QualifierValid=TRUE] Analog Input[" + (index + 1) + "] '" + analogInputsArray[index].Name + "' using qualifier=DigitalInput[" + (analogInputsArray[index].Qualifier) + "]");
                                    SocketIoManager.sockets.emit("AnalogInputHoldTime", { "AnalogInput": index, "HoldTime": "OFF" })
                                    analogInputsArray[index].HoldTimerActive = false;
                                }
                            }
                        }
                    }
                    //#endregion HoldTime logic for Qualifier USED & TRUE
                } else {
                    //#region HoldTime logic for Qualifier USED & FALSE
                    QualifierValid = false;
                    //logger.trace("Analog Inputs - [Qualifier=USED, QualifierValid=FALSE] Analog Input[" + (index+1) + "] using qualifier=DigitalInput[" + (analogInputsArray[index].Qualifier) + "]");

                    if (analogInputsArray[index].HoldTime > 0) {
                        // If there is a hold time ...
                        if (analogInputsArray[index].HoldTimerActive == true) {
                            // Reset hold timer ... it will have to be requalified
                            LogManager.info("Analog Inputs - [HOLD TIMER - HoldTimer=" + (analogInputsArray[index].HoldTime > 0?"RESET (Qualifier Failed)":"NOT_USED") + ", Qualifier=USED, QualifierValid=FALSE] Analog Input[" + (index + 1) + "] '" + analogInputsArray[index].Name + "' using qualifier=DigitalInput[" + (analogInputsArray[index].Qualifier) + "]");
                            SocketIoManager.sockets.emit("AnalogInputHoldTime", { "AnalogInput": index, "HoldTime": "OFF" })
                            analogInputsArray[index].HoldTimerActive = false;
                        }
                    }
                    //#endregion HoldTime logic for Qualifier USED & FALSE
                }
                //#endregion Qualifier USED logic
            } else {
                // Qualifier NOT USED logic
                //#region Qualifier NOT USED logic
                //#region HoldTime logic for QUALIFIER NOT USED
                QualifierValid = true;
                //logger.trace("Analog Inputs - [Qualifier=NOT_USED] Analog Input[" + (index+1) + "]");

                if (analogInputsArray[index].HoldTime > 0) {
                    // If there is a hold time ...
                    if (analogInputsArray[index].HoldTimerActive === false) {
                        // If the hold timer was previously not running ...
                        if (ThresholdChanged) {
                            //  And there was a threshold change ... start the timer
                            analogInputsArray[index].HoldTimer = timeChanged;
                            analogInputsArray[index].HoldTimerActive = true;
                            LogManager.debug("Analog Inputs - [HOLD TIMER - HoldTimer=" + (analogInputsArray[index].HoldTime > 0?"STARTED (Threshold Change)":"NOT_USED") + ", Qualifier=NOT_USED] Analog Input[" + (index + 1) + "] '" + analogInputsArray[index].Name + "'");
                            SocketIoManager.sockets.emit("AnalogInputHoldTime", { "AnalogInput": index, "HoldTime": "HOLDING" })
                        }
                    }
                    else {
                        //  And the threshold has returned to same place ... reset the timer
                        // If the hold timer was previously running ...
                        if (!ThresholdChanged) {
                            if (analogInputsArray[index].HoldTimerActive == true) {
                                LogManager.debug("Analog Inputs - [HOLD TIMER - HoldTimer=" + (analogInputsArray[index].HoldTime > 0?"RESET (Threshold Return)":"NOT_USED") + ", Qualifier=NOT_USED] Analog Input[" + (index + 1) + "] '" + analogInputsArray[index].Name + "'");
                                SocketIoManager.sockets.emit("AnalogInputHoldTime", { "AnalogInput": index, "HoldTime": "OFF" })
                                analogInputsArray[index].HoldTimerActive = false;
                            }
                        }
                    }
                }
                //#endregion HoldTime logic for QUALIFIER NOT USED
                //#endregion Qualifier NOT USED logic
            }
            LogManager.trace("Analog Inputs - readAnalogInputFile: Qualification Logic - QualifierValid=" + QualifierValid);
            //#endregion Qualification logic
            
            // HoldTime logic
            //  - HoldTimeValid = true or false, depending on:
            //      - true: if holdtimer is used & >= holdtime
            //      - true: if holdtimer is not used
            //      - false: all other conditions
            //#region HoldTime logic
            LogManager.trace("Analog Inputs - readAnalogInputFile: HoldTime Logic");
            var HoldTimeValid = false;
            var HoldTimeChanged = 0;
            var HoldTimeActive = false;
            if (analogInputsArray[index].HoldTime > 0) {
                // Using a hold time ...
                if (analogInputsArray[index].HoldTimerActive == true) {
                    // Hold time is valid ...
                    HoldTimeActive = true;
                    HoldTimeChanged = (timeChanged - analogInputsArray[index].HoldTimer) / 1000;
                    if (HoldTimeChanged >= analogInputsArray[index].HoldTime) {
                        HoldTimeValid = true;
                        LogManager.info("Analog Inputs - [HOLD TIMER - HoldTimer=" + Number(analogInputsArray[index].HoldTime).toFixed(2) + "s, Timer=" + Number(HoldTimeChanged).toFixed(2) + "s, HoldTimeValid=TRUE] Analog Input[" + (index+ 1) + "] '" + analogInputsArray[index].Name + "'");
                        // Turn on, then turn off Hold Time button
                        SocketIoManager.sockets.emit("AnalogInputHoldTime", { "AnalogInput": index, "HoldTime": "ON" });
                        (function (index) {
                            setTimeout(function () {
                                SocketIoManager.sockets.emit("AnalogInputHoldTime", { "AnalogInput": index, "HoldTime": "OFF" })
                            }, 1000);
                        })(index);
                    } else {
                        LogManager.info("Analog Inputs - [HOLD TIMER - HoldTimer=" + Number(analogInputsArray[index].HoldTime).toFixed(2) + "s, Timer=" + Number(HoldTimeChanged).toFixed(2) + "s, HoldTimeValid=FALSE] Analog Input[" + (index + 1) + "] '" + analogInputsArray[index].Name + "'");
                        SocketIoManager.sockets.emit("AnalogInputHoldTime", { "AnalogInput": index, "HoldTime": "HOLDING" })
                    }
                }
            } else {
                // Not using a hold time ...
                HoleTimeActive = false;
                HoldTimeValid = true;
                LogManager.trace("Analog Inputs - [HOLD TIMER - HoldTimer=NOT_USED] Analog Input[" + (index + 1) + "] '" + analogInputsArray[index].Name + "'");
            }
            //#endregion HoldTime logic
            
            // Publish logic (real-time info and SNMP traps)
            //  Refer to the 'Crest - Real-time updates and SNMP traps.xlsx' document for cases
            //#region Publish logic
            LogManager.trace("Analog Inputs - readAnalogInputFile: Publish Logic");
            var FormulaIndex;
            if (QualifierValid && HoldTimeValid) {
                // If the change is qualified and holdtime is valid ...
                //#region QualifierValid=TRUE and HoldTimeValid=TRUE preamble
                // Update the local array to the new values
                
                // Calculate Formula info, if present
                analogInputsArray[index].RawValue = readValue;
                analogInputsArray[index].RangedValue = Number(ConvertRawValueToValue(analogInputsArray[index].RawValue)).toFixed(2);// Update the Ranged Value
                //#region Calculate FormulaValue
                FormulaIndex = -1;
                if (analogInputsArray[index].FormulaId !== -1) {
                    // Find formula
                    for (var f = 0; f < FormulasArray.length; f++) {
                        if (analogInputsArray[index].FormulaId === FormulasArray[f].Id) {
                            FormulaIndex = f;
                            break;
                        }
                    }
                    
                    // Check to see if have a formula, if not don't worry about FormulaValue for this change
                    if (FormulaIndex !== -1) {
                        analogInputsArray[index].FormulaValue = _this.CalculateFormula(analogInputsArray[index].RangedValue, FormulasArray[FormulaIndex]);
                    }
                }
                //#endregion Calculate FormulaValue
                //#endregion QualifierValid=TRUE and HoldTimeValid=TRUE preamble
                
                if (ThresholdChanged) {
                    // Threshold change is qualified and hold time is valid, so record the changes
                    //  *** CASE 1 : QualifierValid=TRUE, HoldTimeValid=TRUE, ThresholdChange=TRUE
                    //  *** CASE 2B: QualifierValid=TRUE, HoldTimeValid=TRUE, ThresholdChange=TRUE
                    //  *** CASE 3A: QualifierValid=TRUE, HoldTimeValid=TRUE, ThresholdChange=TRUE
                    //  *** CASE 4D: QualifierValid=TRUE, HoldTimeValid=TRUE, ThresholdChange=TRUE
                    //#region QualifierValid=TRUE, HoldTimeValid=TRUE, ThresholdChange=TRUE
                    analogInputsArray[index].TimeChanged = timeChanged;
                    analogInputsArray[index].ThresholdState = newThreshold;
                    analogInputsArray[index].HoldTimerActive = false;// If the threshold change, set HoldTimerState back to false to restart HoldTimer
                    
                    // Update real-time value, including SNMP because of threshold change
                    //  If this is the first time through, only report out of normal values
                    //  All other times report any changes
                    if ((ANALOG_INPUTS_FIRST_READ == false) || 
                        (ANALOG_INPUTS_FIRST_READ == true && analogInputsArray[index].ThresholdState != 0)) {
                        sendSocketSnmpSmtp(analogInputsArray[index]);                    
                    }
                    
                    // Record the analog input change in the database
                    SqliteManager.run("UPDATE AnalogInputs SET RawValue='" + analogInputsArray[index].RawValue 
                    + "', RangedValue='" + analogInputsArray[index].RangedValue 
                    + "', TimeChanged='" + analogInputsArray[index].TimeChanged 
                    + "', ThresholdState='" + analogInputsArray[index].ThresholdState 
                    + "' WHERE Address = '" + index + "'", 
                    function () { });
                    
                    LogManager.info("Analog Inputs - [REAL-TIME UPDATE - Threshold Changed=TRUE] Analog Input[" + (index + 1) + "] '" + analogInputsArray[index].Name + "' RawValue=" + analogInputsArray[index].RawValue + " RangedValue=" + analogInputsArray[index].RangedValue + "V" + (FormulaIndex !== -1? " FormulaValue=" + analogInputsArray[index].FormulaValue + FormulasArray[FormulaIndex].Units:"") + " Threshold=" + (newThreshold != -1?ThresholdValues[newThreshold]:"UNDEFINED") + " Qualifier=" + (analogInputsArray[index].Qualifier == 0?"NONE":"'Digital Input " + (analogInputsArray[index].Qualifier)) + "' HoldTime=" + (analogInputsArray[index].HoldTime == 0?"0":Number(HoldTimeChanged).toFixed(2)) + "seconds");
                    LogManager.info("Analog Inputs - [THRESHOLD UPDATE - Threshold Changed=TRUE] Analog Input[" + (index + 1) + "] '" + analogInputsArray[index].Name + "' RawValue=" + analogInputsArray[index].RawValue + " RangedValue=" + analogInputsArray[index].RangedValue + "V" + (FormulaIndex !== -1? " FormulaValue=" + analogInputsArray[index].FormulaValue + FormulasArray[FormulaIndex].Units:"") + " Threshold=" + (newThreshold != -1?ThresholdValues[newThreshold]:"UNDEFINED") + " Qualifier=" + (analogInputsArray[index].Qualifier == 0?"NONE":"'Digital Input " + (analogInputsArray[index].Qualifier)) + "' HoldTime=" + (analogInputsArray[index].HoldTime == 0?"0":Number(HoldTimeChanged).toFixed(2)) + "seconds");
                    LogManager.debug("Analog Inputs - [DATABASE UPDATE] Analog Input[" + (index + 1) + "] '" + analogInputsArray[index].Name + "' New Value=" + analogInputsArray[index].RawValue + " New RangedValue=" + analogInputsArray[index].RangedValue + "V New ThresholdState=" + (newThreshold != -1?ThresholdValues[newThreshold]:"UNDEFINED"));
                    LogManager.trace("Analog Inputs - [DATABASE UPDATE] Analog Input[" + (index + 1) + "] =" + JSON.stringify(analogInputsArray[index]));
                    //#endregion QualifierValid=TRUE, HoldTimeValid=TRUE, ThresholdChange=TRUE
                } else {
                    // No threshold changed, so just send real time update
                    //  *** CASE 1 : QualifierValid=TRUE, HoldTimeValid=TRUE, ThresholdChange=FALSE
                    //  *** CASE 2B: QualifierValid=TRUE, HoldTimeValid=TRUE, ThresholdChange=FALSE
                    //  *** CASE 3A: QualifierValid=TRUE, HoldTimeValid=TRUE, ThresholdChange=FALSE
                    //  *** CASE 4D: QualifierValid=TRUE, HoldTimeValid=TRUE, ThresholdChange=FALSE
                    //#region QualifierValid=TRUE, HoldTimeValid=TRUE, ThresholdChange=FALSE
                    
                    // Update real-time value, if there was value change or we're holding
                    if (valueChanged > ANALOG_INPUTS_CHANGE_FILTER) {
                        LogManager.debug("Analog Inputs - [REAL-TIME UPDATE - Threshold Changed=FALSE] Analog Input[" + (index + 1) + "] '" + analogInputsArray[index].Name + "' RawValue=" + analogInputsArray[index].RawValue + " RangedValue=" + analogInputsArray[index].RangedValue + "V" + (FormulaIndex !== -1? " FormulaValue=" + analogInputsArray[index].FormulaValue + FormulasArray[FormulaIndex].Units:""));
                        sendValueAnalogInput("AnalogInput", analogInputsArray[index]);// Q=TRUE, HT=TRUE, No threshold change
                    } 
                    //#endregion QualifierValid=TRUE, HoldTimeValid=TRUE, ThresholdChange=FALSE
                }
            } else {
                
                // If we're qualified and changes are happening within the threshold value, send an update
                if (QualifierValid) {
                    // No threshold changed, so just send real time update
                    //  *** CASE 2A: QualifierValid=TRUE, HoldTimeValid=FALSE, ValueChange=TRUE, ThresholdChange=FALSE
                    //  *** CASE 2A: QualifierValid=TRUE, HoldTimeValid=FALSE, ValueChange=TRUE, ThresholdChange=TRUE
                    //  *** CASE 4C: QualifierValid=TRUE, HoldTimeValid=FALSE, ValueChange=TRUE, ThresholdChange=FALSE
                    //  *** CASE 4C: QualifierValid=TRUE, HoldTimeValid=FALSE, ValueChange=TRUE, ThresholdChange=TRUE
                    //#region QualifierValid=TRUE, HoldTimeValid=FALSE, ValueChange=TRUE, ThresholdChange=DONTCARE
                    analogInputsArray[index].RawValue = readValue;
                    analogInputsArray[index].RangedValue = Number(ConvertRawValueToValue(analogInputsArray[index].RawValue)).toFixed(2);// Update the Ranged Value
                    
                    // Calculate Formula info, if present
                    //#region Calculate FormulaValue
                    FormulaIndex = -1;
                    if (analogInputsArray[index].FormulaId !== -1) {
                        // Find formula
                        for (var f = 0; f < FormulasArray.length; f++) {
                            if (analogInputsArray[index].FormulaId === FormulasArray[f].Id) {
                                FormulaIndex = f;
                                break;
                            }
                        }
                        
                        // Check to see if have a formula, if not don't worry about FormulaValue for this change
                        if (FormulaIndex !== -1) {
                            analogInputsArray[index].FormulaValue = _this.CalculateFormula(analogInputsArray[index].RangedValue, FormulasArray[FormulaIndex]);
                        }
                    }
                    //#endregion Calculate FormulaValue
                    
                    // Update real-time value, if there was value change or we're holding
                    if (valueChanged > ANALOG_INPUTS_CHANGE_FILTER) {
                        LogManager.info("Analog Inputs - [REAL-TIME UPDATE - QualifierValid=" + QualifierValid + ", HoldTimeValid=" + HoldTimeValid + "] Analog Input[" + (index + 1) + "] '" + analogInputsArray[index].Name + "' ValueChanged=" + valueChanged + " RawValue=" + analogInputsArray[index].RawValue + " RangedValue=" + analogInputsArray[index].RangedValue + "V" + (FormulaIndex !== -1? " FormulaValue=" + analogInputsArray[index].FormulaValue + FormulasArray[FormulaIndex].Units:"") + " Threshold=" + (newThreshold != -1?ThresholdValues[newThreshold]:"UNDEFINED"));
                        sendValueAnalogInput("AnalogInput", analogInputsArray[index]);// Q=TRUE, HT=FALSE, No threshold change
                    } 
                    //#endregion QualifierValid=TRUE, HoldTimeValid=FALSE, ValueChange=TRUE, ThresholdChange=DONTCARE
                } else {
                    // Not qualified, don't do anyting ...
                    LogManager.trace("Analog Inputs - [Qualifier=USED, QualifierValid=FALSE] Analog Input[" + (index + 1) + "] '" + analogInputsArray[index].Name + "'");
                }
            }
            //#endregion Publish logic
        }
        catch (error) {
            LogManager.error('Analog Inputs - Process Error (General)=' + error);
            LogManager.error('Analog Inputs - Address=' + (index + 1) + '  previousValue=' + analogInputsArray[index].RawValue + ' readValue=' + readValue);
        }
    }
    else {
    //logger.info("Analog Input[" + (index+1)) + "]= " + analogInputsArray[index].RawValue + " (Disabled)")
    }
}
//#endregion Analog Inputs Processing

//#region Analog Inputs Eventing
function sendSocketSnmpSmtp(analogInput) {
    // SOCKET.IO: Send changed value to all attached clients
    sendValueAnalogInput("AnalogInput", analogInput);// Send Socket, SNMP and SMTP
    
    // SNMP: Send SNMP Trap/Notification if threshold(s) crossed
    sendSNMP(analogInput)

    // SMTP: Send SMTP message if threshold(s) crossed
    sendSMTP(analogInput)
}

function sendValueAnalogInput(dataType, data) {
    //Logger.info("sendValueAnalogInput - DataType='" + dataType + "' Analog Input[" + (data.Address+1) + "]=" + JSON.stringify(data));
    SocketIoManager.sockets.emit(dataType, data);
}

function sendInterimAnalogInputs() {
    //Logger.info("sendValueAnalogInputs - interim update");
    SocketIoManager.sockets.emit("AnalogInputs", { "AnalogInputs": analogInputsArray })
}

// Potential addition to AnalogInputs
//function sendInterimAnalogInputs() {
//    var interimAnalogInputsArray = analogInputsArray;

//    // Check which IO points have a formula
//    for (var i = 0; i < interimAnalogInputsArray.length; i++) {
//        if (interimAnalogInputsArray[i].FormulaId !== -1) {
//            var formulaUnits = "Volts";
//            if (interimAnalogInputsArray[i].FormulaEnabled !== -1) {
//                // If the formula is enabled figure out the units
//                var formulaIndex = -1;
//                for (var f = 0; f < FormulasArray.length; f++) {
//                    if (interimAnalogInputsArray[i].FormulaId === FormulasArray[f].Id) {
//                        formulaIndex = f;
//                        break;
//                    }
//                }

//                if (formulaIndex !== -1 && FormulasArray[formulaIndex].Units.length != 0) {
//                    formulaUnits = FormulasArray[formulaIndex].Units;
//                }
//            }
//        }
//        interimAnalogInputsArray[i].FormulaUnits = formulaUnits;
//    }
//    SocketIoManager.sockets.emit("AnalogInputs", { "AnalogInputs": interimAnalogInputsArray })
//}

function sendValueAnalogInputsConfigured() {
    enabled = 0;
    for (i = 0; i < analogInputsArray.length; i++) {
        if (analogInputsArray[i].Enabled === "TRUE") enabled++;
    }
    SocketIoManager.sockets.emit("AnalogInputsConfigured", { "AnalogInputsTotal": analogInputsArray.length, "AnalogInputsEnabled": enabled })
}

function sendSNMP(analogInput) {
    // SNMP definitions - also refer to the TASC-SUMMIT.MIB file for detailed information
    if (analogInput.Enabled == "FALSE") return;
    if (analogInput.SendEvents == "FALSE") return;
        
    address = parseInt(analogInput.Address) + 1;// To overcome concatenation
    LogManager.trace("Analog Inputs - Sending SNMP Trap for Analog Input[" + address + "] '" + analogInput.Name + "' Threshold=" + analogInput.ThresholdState);
    
    // Calculate Formula info, if present
    var FormulaDefined = (analogInput.FormulaId === -1) ? 0 : 1;// Integer
    var FormulaEnabled = (analogInput.FormulaEnabled === 0) ? 0 : 1;// Integer
    var FormulaValue = "[N/A]";// Octet string - default=""
    var FormulaString = "[N/A]";// Octet string - default="[NONE]"
    var FormulaUnits = "[N/A]";// Octet string - default=""
    
    var FormulaIndex = -1;
    if (tempAnalogInputPoint.FormulaId !== -1) {
        // Find formula
        for (var f = 0; f < FormulasArray.length; f++) {
            if (tempAnalogInputPoint.FormulaId === FormulasArray[f].Id) {
                FormulaIndex = f;
                break;
            }
        }
        
        // If there is a formula for this I/O point, then extract the values
        if (FormulaIndex !== -1) {
            FormulaValue = analogInput.FormulaValue; if (FormulaValue.length == 0) FormulaValue = "[Empty]";
            FormulaUnits = FormulasArray[FormulaIndex].Units; if (FormulaUnits.length == 0) FormulaUnits = "[Empty]";
            FormulaString = FormulasArray[FormulaIndex].Formula; if (FormulaString.length == 0) FormulaString = "[Empty]";
        }
    }
    //logger.debug("Analog Inputs - [SNMP Trap] FormulaDefined='" + FormulaDefined + "'");
    //logger.debug("Analog Inputs - [SNMP Trap] FormulaEnabled='" + FormulaEnabled + "'");
    //logger.debug("Analog Inputs - [SNMP Trap] FormulaValue='" + FormulaValue + "'");
    //logger.debug("Analog Inputs - [SNMP Trap] FormulaUnits='" + FormulaUnits + "'");
    //logger.debug("Analog Inputs - [SNMP Trap] FormulaString='" + FormulaString + "'");
    
    // Define the notification
    var snmpCommand;
    var snmpHost;
    var snmpNotification = "";
    snmpNotification += "1.3.6.1.4.1.11902.3.3.1000.1000.0." + (address - 1 + 1000) + " ";
    snmpNotification += "1.3.6.1.4.1.11902.3.3." + (address) + ".0.1 s '" + moment().toString() + "' ";
    snmpNotification += "1.3.6.1.4.1.11902.3.3." + (address) + ".0.2 i " + (parseInt(analogInput.Address) + 1) + " ";
    snmpNotification += "1.3.6.1.4.1.11902.3.3." + (address) + ".0.3 s '" + analogInput.Name + "' ";
    snmpNotification += "1.3.6.1.4.1.11902.3.3." + (address) + ".0.4 s '" + (Number((ConvertRawValueToValue(analogInput.RawValue))).toFixed(2)).toString() + "' ";
    snmpNotification += "1.3.6.1.4.1.11902.3.3." + (address) + ".0.5 i " + parseInt(analogInput.ThresholdState) + " ";
    snmpNotification += "1.3.6.1.4.1.11902.3.3." + (address) + ".0.6 s '" + analogInput.LowThreshold + "' ";
    snmpNotification += "1.3.6.1.4.1.11902.3.3." + (address) + ".0.7 s '" + analogInput.HighThreshold + "' ";
    snmpNotification += "1.3.6.1.4.1.11902.3.3." + (address) + ".0.8 i " + parseInt(analogInput.HoldTime) + " ";
    snmpNotification += "1.3.6.1.4.1.11902.3.3." + (address) + ".0.9 i " + parseInt(analogInput.Offset * 100) + " ";
    snmpNotification += "1.3.6.1.4.1.11902.3.3." + (address) + ".0.10 i " + parseInt(analogInput.Qualifier) + " ";
    snmpNotification += "1.3.6.1.4.1.11902.3.3." + (address) + ".0.11 i " + parseInt(FormulaDefined) + " ";
    snmpNotification += "1.3.6.1.4.1.11902.3.3." + (address) + ".0.12 i " + parseInt(FormulaEnabled) + " ";
    snmpNotification += "1.3.6.1.4.1.11902.3.3." + (address) + ".0.13 s '" + FormulaValue + "' ";
    snmpNotification += "1.3.6.1.4.1.11902.3.3." + (address) + ".0.14 s '" + FormulaUnits + "' ";
    snmpNotification += "1.3.6.1.4.1.11902.3.3." + (address) + ".0.15 s '" + FormulaString + "' ";
    snmpNotification += "1.3.6.1.4.1.11902.3.3." + (address) + ".0.20 s '" + SummitInfo.MACID + "' ";
    snmpNotification += "1.3.6.1.4.1.11902.3.3." + (address) + ".0.21 s '" + SummitInfo.Name + "' ";

    // Send the SNMP notification
    for (i = 0; i < SnmpManager.length; i++) {
        if (SnmpManager[i].Enabled === "FALSE") continue;

        // Set up the SNMP trap command
        snmpCommand = "snmptrap ";
        if (SnmpManager[i].SnmpVersion == 2) {
            snmpCommand += "-v 2c ";
            snmpCommand += "-c " + SnmpManager[i].Community + " ";
        }
        else if (SnmpManager[i].SnmpVersion == 3) {
            snmpCommand += "-v 3 ";
            var EngineId = "0x" + SnmpManager[i].Community.replace(/\-/g, '');
            snmpCommand += "-e " + EngineId + " ";
            snmpCommand += "-u " + SnmpManager[i].AuthUser + " ";
            if (SnmpManager[i].AuthProtocol > 0) snmpCommand += "-a " + AuthProtocol[SnmpManager[i].AuthProtocol] + " -A " + SnmpManager[i].AuthPassword + " ";
            if (SnmpManager[i].PrivProtocol > 0) snmpCommand += "-x " + PrivProtocol[SnmpManager[i].PrivProtocol] + " -X " + SnmpManager[i].PrivPassword + " ";
            var securityLevel = "-l noAuthNoPriv ";
            if (SnmpManager[i].AuthProtocol > 0 && SnmpManager[i].PrivProtocol == 0)
                securityLevel = "-l authNoPriv ";
            if (SnmpManager[i].AuthProtocol > 0 && SnmpManager[i].PrivProtocol > 0)
                securityLevel = "-l authPriv ";
            snmpCommand += securityLevel;
        } else continue;
        
        // Send trap to host ...
        snmpHost = SnmpManager[i].IpAddress + " '' ";
        LogManager.info("Analog Inputs - Sending SNMP Event to '" + SnmpManager[i].HostName + "' for Analog Input[" + address + "] '" + analogInput.Name + "' Value=" + (Number((ConvertRawValueToValue(analogInput.RawValue))).toFixed(2)).toString() );
        LogManager.trace("Analog Inputs - SendSNMP: " + snmpCommand + snmpHost + snmpNotification);
        exec(snmpCommand + snmpHost + snmpNotification, function (error, stdout, stderr) {
            if (error !== null) {
                LogManager.error("Analog Inputs - SendSNMP. " + error);
            } else {
                LogManager.debug("Analog Inputs - SendSNMP. " + stdout);
            }
        });
    }
}

function sendSMTP(analogInput) {
    // SNMP definitions - also refer to the TASC-SUMMIT.MIB file for detailed information
    if (analogInput.Enabled == "FALSE") return;
    if (analogInput.SendEvents == "FALSE") return;
    
    // Calculate Formula info, if present
    var FormulaDefined = (analogInput.FormulaId === -1) ? 0 : 1;// Integer
    var FormulaEnabled = (analogInput.FormulaEnabled === 0) ? 0 : 1;// Integer
    var FormulaValue = "[N/A]";// Octet string - default=""
    var FormulaString = "[N/A]";// Octet string - default="[NONE]"
    var FormulaUnits = "[N/A]";// Octet string - default=""
    
    var FormulaIndex = -1;
    if (tempAnalogInputPoint.FormulaId !== -1) {
        // Find formula
        for (var f = 0; f < FormulasArray.length; f++) {
            if (tempAnalogInputPoint.FormulaId === FormulasArray[f].Id) {
                FormulaIndex = f;
                break;
            }
        }
        
        // If there is a formula for this I/O point, then extract the values
        if (FormulaIndex !== -1) {
            FormulaValue = analogInput.FormulaValue; if (FormulaValue.length == 0) FormulaValue = "[Empty]";
            FormulaUnits = FormulasArray[FormulaIndex].Units; if (FormulaUnits.length == 0) FormulaUnits = "[Empty]";
            FormulaString = FormulasArray[FormulaIndex].Formula; if (FormulaString.length == 0) FormulaString = "[Empty]";
        }
    }
    
    // Set up the SMTP Email message
    var message = [];
    address = parseInt(analogInput.Address) + 1;// To overcome concatenation
    var event = "'" + analogInput.Name + "' has changed to status=" + ThresholdValues[analogInput.ThresholdState] + " Value="+ ((Number((ConvertRawValueToValue(analogInput.RawValue))).toFixed(2)).toString())+ "V";
    message += moment().toString() + "\n";
    message += "Summit MACID=" + SummitInfo.MACID + "\n";
    message += "Summit Name=" + SummitInfo.Name + "\n";
    message += "Address=" + (parseInt(analogInput.Address) + 1) + "\n";
    message += "Name=" + analogInput.Name + "\n";
    message += "Value=" + ((Number((ConvertRawValueToValue(analogInput.RawValue))).toFixed(2)).toString()) + "\n";
    message += "Status=" + (analogInput.ThresholdState) + "\n";
    message += "Low Threshold=" + (analogInput.LowThreshold) + "\n";
    message += "High Threshold=" + (analogInput.HighThreshold) + "\n";
    message += "Hold Time=" + parseInt(analogInput.HoldTime) + "\n";
    message += "Offset=" + parseInt(analogInput.Offset * 100) + "\n";
    message += "Qualifier=" + parseInt(analogInput.Qualifier) + "\n";
    message += "Formula Defined=" + parseInt(FormulaDefined) + "\n";
    message += "Formula Enabled=" + parseInt(FormulaEnabled) + "\n";
    message += "Formula Value=" + FormulaValue + "\n";
    message += "Formula Units=" + FormulaUnits + "\n";
    message += "Formula=" + FormulaString + "\n";
    
    var ioInfo = "Analog Input[" + address + "] '" + analogInput.Name + "'";
    
    LogManager.info("Analog Inputs - Sending SMTP Event for Analog Input[" + address + "] '" + analogInput.Name + "'");
    SmtpManager.SendSmtp(message, ioInfo, event, function (result) {
        LogManager.trace("Analog Inputs - SMTP Email Event Response for Analog Input[" + address + "]: " + result.result);
    })
}
//#endregion Analog Inputs Send and SNMP

//#region Analog Inputs Formulas Initialization
exports.InitializeFormulas = function (db, logger, io, callback) {
    SqliteManager = db;
    LogManager = logger;
    SocketIoManager = io;
    
    // Try getting the Formulas table
    LogManager.info('Formulas.InitializeFormulas: Initializing formula database table');
    SqliteManager.get("SELECT name FROM sqlite_master WHERE type='table' AND name='Formulas'", function (err, rows) {
        if (err !== null) {
            LogManager.error('Formulas.InitializeFormulas - Initialize Error (Select)=' + JSON.stringify(err));
        }
        else if (typeof rows === "undefined") {
            // We read nothing, so let's create the table ...
            LogManager.info("Formulas.InitializeFormulas - Initialize by creating database table ...");
            SqliteManager.exec("CREATE TABLE if not exists Formulas (Name TEXT, Formula TEXT, Units TEXT, Decimals INTEGER, TimeChanged TEXT )", function (err) {
                if (err !== null) {
                    LogManager.error('Formulas.InitializeFormulas - Initialize Error (Create)=' + JSON.stringify(err));
                }
                else {
                    if (typeof callback == "function") {
                        LogManager.info('Formulas.InitializeFormulas - Initialized new database table');
                        callback(FormulasArray);
                    }
                }
            });
        }
        else {
            // Read from the database and populate the Formulas array
            LogManager.info("Formulas.InitializeFormulas - Read Formulas from database ...");
            numberOfFormulas = 0;
            FormulasArray = [];
            SqliteManager.each("SELECT rowid AS Id, Name, Formula, Units, Decimals, TimeChanged FROM Formulas", function (err, row) {
                if (err !== null) {
                    LogManager.error('Formulas.InitializeFormulas - Initialize Error (Select)=' + JSON.stringify(err));
                } else {
                    // Copy each row into Formulas array
                    LogManager.trace('Formulas.InitializeFormulas - Processing Host[' + (numberOfFormulas + 1) + ']=' + row.Name);
                    FormulasArray.push({
                        Id : row.Id,
                        Name : row.Name,
                        Formula : row.Formula,
                        Units: row.Units,
                        Decimals: row.Decimals,
                        TimeChanged : row.TimeChanged
                    });
                    LogManager.info("Formulas.InitializeFormulas - Read Formulas from Database[" + numberOfFormulas + "]=" + JSON.stringify(FormulasArray[numberOfFormulas]));
                    numberOfFormulas++;
                }
            }, function () {
                // Retrurn the complete list of Formulas
                LogManager.info('Formulas.InitializeFormulas - Initialized from existing database table');
                if (typeof callback == "function") {
                    callback(FormulasArray);
                }
            });
        }
    });
}

exports.UpdateFormulasPorts = function (formulas) {
    FormulasArray = formulas;
    LogManager.Info("Formulas.UpdateAccesControlPorts: formulas=" + JSON.stringify(formulas));
}
//#endregion Analog Inputs Formulas Initialization

//#region Analog Inputs Formulas API

// Library to manage parsing and evaluation of math expressions within Crest formulas
var mathparser = require('node-expression-eval');

// Get Formulas
exports.GetFormulas = function (callback) {
    
    // Get the FormulasArray by populating latest from database
    FormulasArray = [];
    SqliteManager.all("SELECT rowid as Id, Name, Formula, Units, Decimals, TimeChanged FROM Formulas ORDER BY Id", function (err, rows) {
        if (err) {
            throw LogManager.error('Formulas.GetFormulas: SELECT users - Error (Read)=' + err);
        }
        
        // Repopulate AccessControlUsers array        
        FormulasArray = [];
        for (i = 0; i < rows.length; i++) {
            FormulasArray.push({
                "Id"            : rows[i].Id,
                "Name"          : rows[i].Name,
                "Formula"       : rows[i].Formula,
                "Units"         : rows[i].Units,
                "Decimals"      : rows[i].Decimals,
                "TimeChanged"   : rows[i].TimeChanged
            });
        }
        LogManager.debug('Formulas.GetFormulas: Users=' + JSON.stringify(FormulasArray));
        
        callback(FormulasArray);
    });
};

// Add Formula
exports.AddFormula = function (req, callback) {
    
    // Add Formula
    var TimeChanged = moment();
    var dbcommand = "'" + req.body.Name + "', '" + req.body.Formula + "', '" + req.body.Units + "', " + req.body.Decimals + ", '" + TimeChanged + "'";
    LogManager.debug("Formulas.AddFormula: Add statement=" + dbcommand);
    LogManager.info("Formulas.AddFormula: Formula name=" + req.body.Name);
    SqliteManager.run("INSERT INTO Formulas (Name, Formula, Units, Decimals, TimeChanged) VALUES (" + dbcommand + ")", function (err) {
        if (err !== null) {
            throw LogManager.error('Formulas.AddFormula: Formula=' + req.body.Name + ' - Error (Update)=' + JSON.stringify(err));
        }
        
        _this.GetFormulas(function (FormulasArray) {
            callback(FormulasArray)
        });
    });
};

// Return Units
exports.GetAnalogInputUnits = function (Address) {
    var FormulaIndex = -1;
    if (analogInputsArray[Address].FormulaId !== -1) {
        // Find formula
        for (var f = 0; f < FormulasArray.length; f++) {
            if (analogInputsArray[Address].FormulaId === FormulasArray[f].Id) {
                FormulaIndex = f;
                break;
            }
        }

        var FormulaUnits = FormulasArray[FormulaIndex].Units;
        if (FormulaUnits.length == 0)
            return ""
        else
            return FormulaUnits;

    } else {
        return "Volts";
    }
}

// Return FormulaValue
exports.GetFormulaValue = function (Address) {
    return analogInputsArray[Address].FormulaValue;
}

// Return FormulaLowThreshold
exports.GetFormulaLowThreshold = function (Address) {
    return analogInputsArray[Address].FormulaLowThreshold;
}

// Return FormulaHighThreshold
exports.GetFormulaHighThreshold = function (Address) {
    return analogInputsArray[Address].FormulaHighThreshold;
}

// Delete an existing Formula
exports.DeleteFormula = function (req, callback) {
    
    // Delete Formula
    var dbcommand = "DELETE FROM Formulas WHERE rowid='" + req.body.Id + "'";
    LogManager.debug("Formulas.DeleteFormula: Delete statement=" + dbcommand);
    LogManager.info("Formulas.DeleteFormula: Formula name=" + req.body.Name);
    SqliteManager.run(dbcommand, function (err) {
        if (err !== null) {
            throw LogManager.error('Formulas.DeleteFormula: User=' + req.params.formula + ' - Error (Update)=' + JSON.stringify(err));
        }
        
        _this.GetFormulas(function (Formulas) {
            callback(Formulas)
        });
    });

};

// Modify an existing Formula
exports.ModifyFormula = function (req, callback) {
    
    // Modify Formula
    var TimeChanged = moment();
    var dbcommand = "UPDATE Formulas SET Name='" + req.body.Name + "', Formula='" + req.body.Formula + "', Units='" + req.body.Units + "', Decimals=" + req.body.Decimals + ", TimeChanged='" + TimeChanged + "' WHERE rowid='" + req.body.Id + "'";
    LogManager.debug("Formulas.ModifyFormula: Modify statement=" + dbcommand);
    LogManager.info("Formulas.ModifyFormula: Formula name=" + req.body.Name);
    SqliteManager.run(dbcommand, function (err) {
        if (err !== null) {
            throw LogManager.error('Formulas.ModifyFormula: Formula=' + req.body.formula + ' - Error (Update)=' + JSON.stringify(err));
        }
        
        _this.GetFormulas(function (FormulasArray) {
            callback(FormulasArray)
        });
    });
};

exports.ValidateFormula = function (req, callback) {
    
    // Validate the formula
    var validformula = { result: "UNKNOWN" };
    try {
        var expression = mathparser.parse(req.body.Formula);
        LogManager.info("Formulas.ValidateFormula: Successful parse of '" + req.body.Formula + "'");
        validformula = { result: "TRUE" };// Got here, so it's good!
    }
    catch (err) {
        LogManager.error("Formulas.ValidateFormula: Unsuccessful parse of '" + req.body.Formula + "' err=" + err);
        validformula = { result : "FALSE" };// Something happened during the parsing
    } finally {
        callback(validformula);
    }
};

exports.EvaluateFormula = function (formula, callback) {
    LogManager.info("Formulas.EvaluateFormula: formula information=" + JSON.stringify(formula));
    var result = _this.CalculateFormula(formula.TestValue, formula);
    LogManager.info("Formulas.EvaluateFormula: result=" + result);
    callback({ result: result });
}

// Evaluate an existing Formula
exports.CalculateFormula = function (rawValue, formula) {
    
    // Evaluate Formula
    var calculatedResult = "ERROR";
    try {
        var expression = mathparser.parse(formula.Formula);
        //logger.info("Formulas.CalculateFormula: Successful parse of '" + formula.Formula + "' expression='" + expression + "' rawValue=" + rawValue);
        
        var rawResult = expression.evaluate({ x: rawValue });
        //logger.info("Formulas.CalculateFormula: Successful evaluation - result= '" + rawResult + "'");
        
        calculatedResult = Number(rawResult).toFixed(formula.Decimals);
    }
    catch (err) {
        LogManager.error("Formulas.CalculateFormula: Unsuccessful evaluation of '" + formula.Formula + "' err=" + err);
        calculatedResult = "ERROR";// Something happened during the parsing
    } finally {
        return calculatedResult;
    }
};

exports.UpdateSnmpManager = function (snmp) {
    SnmpManager = snmp;
}

exports.UpdateSummitInfo = function (info) {
    SummitInfo = info;
}
//#endregion Analog Inputs Formulas API
